#! /usr/bin/perl
# 
#    Copyright 2006 Cybozu Labs, Inc.
#    Coyprgiht 2004,2005 Kazuho Oku
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

use strict;
use warnings;
use CGI;
use Cwd;
use Config;
use POSIX qw(_exit);
use IO::Select;
use IO::Socket;


# parameters

my $SCRIPT_FILENAME =
    shift(@ARGV) || $ENV{SCRIPT_FILENAME} || die "No script filename";
$SCRIPT_FILENAME = cwd() . "/$SCRIPT_FILENAME" if $SCRIPT_FILENAME !~ m|^/|;
my $PORT = $ENV{WEBDB_PORT} || die "WEBDB_PORT not set\n";
my $TIMEOUT = $ENV{WEBDB_TIMEOUT} || 30;

# globals

my $listener;
my $client_sock;
my $dbg_sock;
my %state = (terminated => 0);
my $root_package;
my $q; # CGI object


# setup

my $cleanup_done = 0;

sub cleanup {
    if ($cleanup_done) {
	return;
    }
    $cleanup_done = 1;
    if (defined($dbg_sock) && $dbg_sock->opened()) {
	$dbg_sock->write("q\n");
	$dbg_sock->close();
    }
}

END {
    cleanup();
}

sub on_exit {
    my ($sig_name) = @_;
    cleanup();
    my %sig_num;
    my @names = split(' ', $Config{sig_name});
    @sig_num{@names} = split(' ', $Config{sig_num});
    exit($sig_num{$sig_name} + 128);
}

$SIG{PIPE} = 'IGNORE';
$SIG{ALRM} = $SIG{HUP} = $SIG{INT} = $SIG{PIPE} = $SIG{QUIT} = $SIG{TERM} = \&on_exit;

# prepare listener for debugger
my $dbg_port;
my $dbg_listener;
for (my $i = 0; $i < 100; $i++) {
    $dbg_port = 10000 + int(rand(10000));
    $dbg_listener = IO::Socket::INET->new(Listen => SOMAXCONN,
					  LocalAddr => '127.0.0.1',
					  LocalPort => $dbg_port,
					  Proto => 'tcp',
					  ReuseAddr => 1);
    if ($dbg_listener) {
	last;
    }
}
unless ($dbg_listener) {
    die "Failed to allocate a TCP port\n";
}

print STDOUT "$dbg_port\n";
STDOUT->close();

# wait until the debugger connects
$dbg_sock = $dbg_listener->accept();
unless ($dbg_sock) {
    die "debugger connection timeout\n";
}
$dbg_listener->close();
undef $dbg_listener;

# get first prompt
read_debugger();

# start listener
for (my $i = 0; $i < $TIMEOUT; $i++) {
    $listener = IO::Socket::INET->new(Listen => SOMAXCONN,
				      LocalAddr => '127.0.0.1',
				      LocalPort => $PORT,
				      Proto => 'tcp',
				      ReuseAddr => 1);
    if ($listener) {
	last;
    }
    sleep(1);
}
unless ($listener) {
    die "failed to setup listener socket";
}

# get root package
$dbg_sock->print('print ${DB::OUT} ${DB::package}' . "\n");
$root_package = read_debugger();
chomp($root_package);

# main loop
while (1) {
    $client_sock = accept_timeout();
    unless ($client_sock) {
	exit 0;
    }
    handle_connection();
    $client_sock->close();
    undef $client_sock;
    if ($state{terminated}) {
        last;
    }
}
cleanup();
_exit(0);

sub accept_timeout {
    if ($TIMEOUT == 0) {
	return $listener->accept();
    }
    
    my $wait_until = time() + $TIMEOUT;
    while (1) {
	my $timeleft = $wait_until - time();
	if ($timeleft <= 0) {
	    last;
	}
	alarm($timeleft);
	my $sock = $listener->accept();
	alarm(0);
	return $sock if $sock;
    }
    return undef;
}

sub normalize_path {
    my ($path) = @_;
    $path =~ s|/{2,}|/|g;
    $path =~ s|/\./|/|g;
    while ($path =~ s|^/\.\./|/|) {
    }
    $path =~ s|/[^/]+/\.\./|/|g;
    return $path;
}

sub read_debugger {
    my ($pwd) = @_;
    my $response = '';
    while (1) {
	my $buf;
	my $len = $dbg_sock->sysread($buf, 1000000);
	unless ($len) {
	    # debugger closed unexpectedly	    
	    $state{terminated} = 1;
	    last;
	}
	my (@lines) = split(/\n/, $buf);
	foreach my $l (@lines) {
	    if ($l =~ /^[^ \t:\(]*::[^ \t\(]*\((.+):(\d+)\):/) {
		$state{file} = $1;
		$state{line} = $2;
		while ($state{file} =~ /\[([^:]+)\:(\d+)\]$/) {
		    $state{file} = $1;
		    $state{line} = $2;
		}
	    } elsif ($l =~ /^Debugged program terminated/) {
		$state{terminated} = 1;
		$dbg_sock->print("q\n");
		$dbg_sock->close();
		last;
	    }
	}
	my $got_cursor = 0;
	if (scalar(@lines) && $lines[$#lines] =~ /^  DB<\d+> $/) {
	    $got_cursor = 1;
	    pop(@lines);
	}
	if (scalar(@lines)) {
	    $response .= join("\n", @lines);
	    if ($got_cursor <=> 1 && $buf =~ /\n$/) {
		$response .= "\n";
	    }
	}
	if ($got_cursor) {
	    last;
	}
    }
    if (IO::Select->new(\*STDIN)->can_read(0)) {
	# read from stderr of perl
	my $buf;
	my $len = sysread(STDIN, $buf, 10000);
	if ($len <=> 0) {
	    $state{stderr} .= $buf;
	    $state{got_stderr} = 1;
	    print STDERR $buf;
	    flush STDERR;
	    unless (defined($state{file})) {
		# no filename, bug got stderr => is compilation error
		if ($buf =~ / at ([^ ]*) line (\d+)[,.\n]/) {
		    $state{file} = $1;
		    $state{line} = $2;
		}
	    }
	} else {
	    # stderr of perl closed
	    $state{terminated} = 1;
	}
    }
    if ($state{file} !~ m|^/|) {
	if ($pwd ne '') {
	    $state{file} = normalize_path("$pwd/$state{file}");
	} else {
	    $dbg_sock->print("print STDERR `pwd`;\n");
	    $pwd = <STDIN>;
	    unless (chomp($pwd)) {
		print STDERR "unexpected response from perl debugger.\n";
		$dbg_sock->print("q\n");
		$state{terminated} = 1;
	    } else {
		read_debugger($pwd);
	    }
	}
    }
    return $response;
}

sub get_files_loaded {
    $dbg_sock->print('print ${DB::OUT} join("\n", values(%main::INC))."\n"' .
		     "\n");
    my $files_str = read_debugger();
    my @files;
    my $pwd;
    push(@files, $SCRIPT_FILENAME);
    foreach my $file (split("\n", $files_str)) {
	if ($file ne '') {
	    if ($file !~ m|^/|) {
		unless ($pwd) {
		    $dbg_sock->print('print ${DB::OUT} `pwd`' . "\n");
		    $pwd = read_debugger();
		    print STDERR "PWD: \"$pwd\"\n";
		}
		$file = normalize_path(cwd() ."$file");
	    }
	    push(@files, $file);
	}
    }
    return @files;
}

sub set_breakpoint {
    my ($file, $line, $cond) = @_;
    my $err;
    
    $dbg_sock->printf("f %s\n", $file);
    read_debugger();
    
    if (defined($cond)) {
	$dbg_sock->printf("b %d %s\n", $line, $cond);
    } else {
	$dbg_sock->printf("B %d\n", $line);
    }
    $err = read_debugger();
    return if $err;
    return 1;
}

sub handle_connection {
    # read request
    my $req = $client_sock->getline();
    if ($req !~ /^GET (\S+) HTTP\/1.\d+\r\n$/) {
	write_http_error($client_sock, 400, 'Bad Request');
	return;
    }
    my $path = $1;
    # skip headers
    while (1) {
	my $l = $client_sock->getline();
	unless ($l =~ s/[\r\n]+$//) {
	    write_http_error($client_sock, 400, 'Bad Request');
	    return;
	}
	if ($l eq '') {
	    last;
	}
    }
    # handle request
    $q = new CGI($path =~ /\?/ ? $' : '');
    select $client_sock;
    print "HTTP/1.0 200 OK\r\n";
    handle_request($q);
    select STDOUT;
    return;
}

sub dump_stacktrace {
    $state{stack} = {};
    $dbg_sock->print("T\n");
    my $res = read_debugger();
    my $level = 0;
    foreach my $line (split("\n", $res)) {
	if ($line =~ /^.\s*=\s*([^ ]*)::([^ :\(]*)/) {
	    my $frame = {
		'package' => $1,
		'function' => $2
		};
	    if ($line =~ / called from file \`([^\']*)\' line (\d+)/) {
		$frame->{caller_file} = $1;
		$frame->{caller_line} = $2;
	    }
	    if ($level == 0) {
		$frame->{file} = $state{file};
		$frame->{line} = $state{line};
	    } else {
		$frame->{file} =
		    $state{stack}->{$level - 1}->{caller_file};
		$frame->{line} =
		    $state{stack}->{$level - 1}->{caller_line};
	    }
	    $state{stack}->{$level} = $frame;
	    $level++;
	}
    }
    if ($level == 0) {
	$state{stack}->{0} = {
	    package => $root_package,
	    file => $state{file},
	    line => $state{line}
	};
    } else {
	$state{stack}->{$level} = {
	    package => $root_package,
	    file => $state{stack}->{$level - 1}->{caller_file},
	    line => $state{stack}->{$level - 1}->{caller_line}
	};
    }
}

sub dump_subvalue {
    my ($container, $n, $v, $lines) = @_;
    my $line;
    if ($v eq '(') {
	$container->{$n} = {};
	while (scalar(@$lines)) {
	    $line = shift(@$lines);
	    if ($line =~ /^\s*\)/) {
		last;
	    }
	    if ($n =~ /^\@/ ?
		$line =~ /^\s+(\d+)\s+/ :
		($line =~ /^\s+(\d+)\s*=>\s*/ ||
		 $line =~ /^\s+\'([^\']*)\'\s*=>\s*/)) {
		dump_subvalue($container->{$n}, $1, $', $lines);
	    }
	}
    } elsif ($v =~ /(ARRAY|HASH)\(0x[0-9A-Fa-f]+\)$/) {
	my $type = $1;
	$container->{$n} = {};
	if (scalar(@$lines) && $lines->[0] =~ /^(\s+)/) {
	    my $indent = $1;
	    while (scalar(@$lines)) {
		if ($lines->[0] !~ /^$indent/) {
		    last;
		}
		$line = shift(@$lines);
		if ($type eq 'ARRAY' ?
		    $line =~ /^\s+(\d+)\s+/ :
		    ($line =~ /^\s+(\d+)\s*=>\s*/ ||
		     $line =~ /^\s+\'([^\']*)\'\s*=>\s*/)) {
		    dump_subvalue($container->{$n}, $1, $', $lines);
		}
	    }
	}
    } elsif ($v =~ /^SCALAR\(0x[0-9A-Fa-f]+\)$/) {
	$container->{$n} = {};
	if (scalar(@$lines) && shift(@$lines) =~ /^\s+->\s*/) {
	    dump_subvalue($container->{$n}, undef, $', $lines);
	}
    } else {
	$container->{$n} = $v;
    }
}

sub dump_vars {
    my ($container, $cmd) = @_;
    my $res;
    $dbg_sock->printf("$cmd\n");
    $res = read_debugger();
    unless ($res =~ /^[\$\@\%].*\s*=/) {
	return;
    }
    my @lines = split("\n", $res);
    while (scalar(@lines)) {
	my $line = shift(@lines);
	if ($line =~ /^([\$\@\%]\S*)\s*=\s*/) {
	    dump_subvalue($container, $1, $', \@lines);
	}
    }
}

sub dump_lexical_vars {
    my ($level) = @_;
    dump_vars($state{var}->{lexical} = {}, "y $level");
}

sub dump_package_vars {
    my ($package) = @_;
    dump_vars($state{var}->{package} = {}, "V $package");
}

sub dump_breakpoints {
    my $h = $state{breakpoints} = {};
    $dbg_sock->printf("L\n");
    my $res = read_debugger();
    my @lines = split("\n", $res);
    my $cur_file;
    my $cur_line;
    foreach my $line (@lines) {
	if ($line =~ /^(\S.*):/) {
	    $cur_file = $1;
	} elsif ($line =~ /^ (\d+):/) {
	    $cur_line = $1;
	} elsif ($line =~ /^\s+break\s+if\s+\(1\)/) {
	    if (defined($cur_file) && defined($cur_line)) {
		$h->{$cur_file} = {} unless defined($h->{$cur_file});
		$h->{$cur_file}->{$cur_line} = 1;
		undef $cur_line;
	    }
	}
    }
}

sub set_value {
    my ($path, $value) = @_;
    my @path = split('/', $path);
    shift(@path); # strip ^(package|lexical)/
    my $expr = <<'EOF';
local $sv = sub {
    my ($o, $v, @n) = @_;
    if (scalar(@n) == 0) {
	$$o = $v;
    } else {
	my $s = shift(@n);
	if (ref($$o) eq 'HASH') {
	    $sv->(\$$o->{$s}, $v, @n);
	} elsif (ref($$o) eq 'ARRAY') {
	    $sv->(\$$o->[$s], $v, @n);
	} elsif (ref($$o) eq 'SCALAR') {
	    $sv->($$o, $v, @n);
	}
    }
};
$sv->(%s, %s, (%s));
EOF
    ;
    $expr =~ s/\n//g;
    my $base = shift(@path);
    if ($base =~ /^[\@\%]/) {
	$base = "\\\\$base";
    } else {
	$base = "\\$base";
    }
    my $path_expr = join("','", @path);
    if ($path_expr ne '') {
	$path_expr = "'$path_expr'";
    }
    $dbg_sock->printf("$expr\n", $base, $value, $path_expr);
    print STDERR read_debugger();
}

sub handle_request {
    $state{got_stderr} = 0;
    my $cmd = $q->param('cmd') || '';
    foreach my $value ($q->param('set_value')) {
	if ($value ne '') {
	    my ($p, $v) = split('=', $value, 2);
	    set_value($p, $v);
	}
    }
    if ($cmd eq 'n' || $cmd eq 's') {
	my $old_fileline = "$state{file}:$state{line}";
	while ($state{terminated} == 0 &&
	       $old_fileline eq "$state{file}:$state{line}") {
	    $dbg_sock->print("$cmd\n");
	    read_debugger();
	}
    } elsif ($cmd eq 'q') {
	$dbg_sock->print("q\n");
	$dbg_sock->close();
    } elsif ($cmd eq 'set_breakpoint') {
	my $file = $q->param('file');
	my $line = $q->param('line');
	my $cond = $q->param('cond');
	if (defined($file) && defined($line) && defined($cond)) {
	    set_breakpoint($file, $line, $cond);
	}
    } elsif ($cmd ne '') {
	$dbg_sock->print("$cmd\n");
	read_debugger();
    }
    dump_stacktrace();
    $state{var} = {};
    my $stacklevel = ($q->param('stacklevel') || 0) + 0;
    unless (defined($state{stack}->{$stacklevel})) {
	$stacklevel = 0;
    }
    dump_package_vars($state{stack}->{$stacklevel}->{package});
    dump_lexical_vars($stacklevel);
    $state{breakpoints} = {};
    dump_breakpoints();
    send_response();
    if ($cmd eq 'q') {
	exit(0);
    }
}

sub send_hash {
    my ($h) = @_;
    foreach my $n (sort(keys(%{$h}))) {
	my $v = $h->{$n};
	$n =~ s/\W/sprintf("%%%02x", ord($&))/eg;
	if (ref($v) eq 'HASH') {
	    print "$n:(\n";
	    send_hash($v);
	    print ")\n";
	} else {
	    $v =~ s/\W/sprintf("%%%02x", ord($&))/eg;
	    print "$n:$v\n";
	}
    }
}

sub send_response {
    print $q->header(-type => 'text/plain');
    send_hash(\%state);
}

sub write_http_error {
    my ($sock, $st_code, $st_text, $type, $content) = @_;
    unless (defined($type)) {
	$type = 'text/plain';
    }
    unless (defined($content)) {
	$content = $st_text;
    }
    $sock->print("HTTP/1.0 $st_code $st_text\r\n" .
		 "Content-Type: $type\r\n" .
		 "\r\n" .
		 $content);
}

1;
