#! /usr/bin/perl

# $Id: MiniTmpl.pm,v 1.5 2005/02/19 13:57:30 kazuho Exp $
#    Copyright 2004 Kazuho Oku
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


package MiniTmpl;

use strict;
use CGI;
use Carp;

use vars qw(@EXPORT @EXPORT_OK %EXPORT_TAGS @ISA $VERSION);

require Exporter;
@ISA = qw(Exporter);
$VERSION = '0.01';
@EXPORT = qw();
@EXPORT_OK = qw(js_escape uri_escape uri_unescape);
%EXPORT_TAGS = (all => [qw(js_escape uri_escape uri_unescape)]);


sub new {
    my ($class, @initializer) = @_;
    my $self = { @initializer };
    bless $self, ref $class || $class;
    unless (defined($self->{'template'})) {
	$self->load();
    }
    unless (ref($self->{'param'}) eq 'HASH') {
	$self->{'param'} = {};
    }
    return $self;
}

sub load {
    my $self = shift;
    my $fh;
    open($fh, $self->{'file'}) ||
	croak("failed to open file: $self->{'file'}");
    my $lineno = 1;
    while (my $line = <$fh>) {
	$line =~ s/<\%/$&$lineno /g;
	$self->{'template'} .= $line;
	$lineno++;
    }
    close($fh);
}

sub param {
    my $self = shift;
    if ($#_ == 0) {
	return $self->{'param'}->{$_[0]};
    } else {
	while ($#_ >= 0) {
	    my $n = shift;
	    my $v = shift;
	    $self->{'param'}->{$n} = $v;
	}
    }
    return $self->{'param'};
}

sub output {
    my $self = shift;
    my $template = $self->{'template'};
    my @stack;
    my $output = '';
    unshift(@stack, { param => $self->{'param'}, show => 1 });
    while ($template =~ /<\%(\d+)\s*([^>]+)\%>/) {
	if ($stack[0]->{'show'}) {
	    $output .= $`;
	}
	$template = $';
	my ($lineno, $command) = ($1, $2);
	$command =~ s/^\s*//;
	$command =~ s/\s*$//;
	if ($command =~ /^print\s+/) {
	    if ($stack[0]->{'show'}) {
		my $mode = '';
		my $name = $';
		if ($name =~ /\s+/) {
		    $mode = $`;
		    $name = $';
		}
		my $separator = '';
		if ($name =~ /\s+/) {
		    $separator = $`;
		    $name = $';
		}
		my $value = $stack[0]->{'param'}->{$name};
		if (ref($value) eq 'ARRAY') {
		    for (my $i = 0; $i <= $#$value; $i++) {
			$output .= $self->_output_value($mode, $value->[$i]);
			unless ($i == $#$value) {
			    $output .= CGI::escapeHTML($separator);
			}
		    }
		} else {
		    $output .= $self->_output_value($mode, $value);
		}
	    }
	} elsif ($command =~ /^if\s+/) {
	    my $name = $';
	    unshift(@stack, { command => $command,
			      start_lineno => $lineno,
			      param => $stack[0]->{'param'} });
	    if ($stack[0]->{'param'}->{$name}) {
		$stack[0]->{'show'} = $stack[1]->{'show'};
		$stack[0]->{'if_shown'} = 1;
	    }
	} elsif ($command =~ /^unless\s+/) {
	    my $name = $';
	    unshift(@stack, { command => $command,
			      start_lineno => $lineno,
			      param => $stack[0]->{'param'} });
	    unless ($stack[0]->{'param'}->{$name}) {
		$stack[0]->{'show'} = $stack[1]->{'show'};
		$stack[0]->{'if_shown'} = 1;
	    }
	} elsif ($command =~ /^elsif\s+/) {
	    my $name = $';
	    $stack[0]->{'command'} = $command;
	    $stack[0]->{'start_lineno'} = $lineno;
	    if (! $stack[0]->{'if_shown'} && $stack[0]->{'param'}->{$name}) {
		$stack[0]->{'show'} = $stack[1]->{'show'};
		$stack[0]->{'if_shown'} = 1;
	    } else {
		$stack[0]->{'show'} = 0;
	    }
	} elsif ($command eq 'else') {
	    $stack[0]->{'command'} = $command;
	    $stack[0]->{'start_lineno'} = $lineno;
	    if (! $stack[0]->{'if_shown'}) {
		$stack[0]->{'show'} = $stack[1]->{'show'};
		$stack[0]->{'if_shown'} = 1;
	    } else {
		$stack[0]->{'show'} = 0;
	    }
	} elsif ($command eq 'endif') {
	    croak "unexpected `endif' at line $lineno" if $#stack < 0;
	    shift(@stack);
	} elsif ($command =~ /^foreach\s+/) {
	    my $name = $';
	    my $list = $stack[0]->{'param'}->{$name};
	    if ($stack[0]->{'show'} &&
		ref $list eq 'ARRAY' && $#$list <=> -1) {
		my ($param, @leftover) = @$list;
		unshift(@stack, { param => $param,
				  loop_data => \@leftover,
				  show => 1,
				  template => $template });
	    } else {
		unshift(@stack, { param => {}, show => 0, loop_data => [] });
	    }
	    $stack[0]->{'command'} = $command;
	    $stack[0]->{'start_lineno'} = $lineno;
	} elsif ($command eq 'end') {
	    croak "unexpected `end' at line $lineno"
		unless defined($stack[0]->{'loop_data'});
	    if ($#{$stack[0]->{'loop_data'}} <=> -1) {
		$stack[0]->{'param'} = shift(@{$stack[0]->{'loop_data'}});
		$template = $stack[0]->{'template'};
	    } else {
		shift(@stack);
	    }
	} else {
	    croak "unexpected command `$command' at line $lineno";
	}
    }
    croak "`$stack[0]->{'command'}' starting at line" .
	" $stack[0]->{'start_lineno'} not closed" if ($#stack <=> 0);
    $output .= $template;
    return $output;
}

sub _output_value {
    my ($self, $mode, $value) = @_;
    my $output;
    if ($mode eq '' || $mode eq 'escape') {
	$output = CGI::escapeHTML($value);
    } elsif ($mode eq 'raw') {
	$output = $value;
    } elsif ($mode eq 'js_dblquote') {
	$output = '"' . js_escape($value) . '"';
    } elsif ($mode eq 'js_escape') {
	$output = js_escape($value);
    } elsif ($mode eq 'js_quote') {
	$output = '\'' . js_escape($value) . '\'';
    } elsif ($mode eq 'ucfirst') {
	$output = CGI::escapeHTML(ucfirst($value));
    } elsif ($mode eq 'uri_escape') {
	$output = uri_escape($value);
    } elsif ($mode =~ /^\%/) {
	$output = CGI::escapeHTML(sprintf($mode, $value));
    } else {
	croak("unexpected print mode $mode");
    }
    return $output;
}

sub js_escape {
    shift if (ref $_[0]);
    my $str = $_[0];
    $str =~ s/[\0-\037\177-\377\'\"]/sprintf('\x%02x', ord($&))/eg;
    return $str;
}

sub uri_escape {
    shift if (ref $_[0]);
    my $str = $_[0];
    $str =~ s/[^A-Za-z0-9\-_.!~*'()]/sprintf('%%%02x', ord($&))/eg;
    return $str;
}

sub uri_unescape {
    shift if (ref $_[0]);
    my $str = $_[0];
    $str =~ s/\+/ /g;
    $str =~ s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg;
    return $str;
}

1;
