#! /usr/bin/perl
# 
#    Copyright 2006 Cybozu Labs, Inc.
#    Coypright 2004,2005 Kazuho Oku
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

use strict;
use CGI qw/:standard Vars/;
use File::Basename;
use IO::Socket::INET;
use locale;
use lib '../lib';
use MiniTmpl qw(:all);


sub sigusr1_handler {
    # just receive the signal
}

#CGI::save(*STDERR);

$SIG{USR1} = \&sigusr1_handler;

my $cmd = param('cmd');
if ($cmd eq '') {
    foreach my $k (Vars()) {
	if ($k =~ /^cmd-(.*)\.x$/) {
	    $cmd = $1;
	    last;
	}
    }
}
my $invisible_var_type = param('invisible_var_type');
my $stacklevel = param('stacklevel') + 0;

print header(-cache_control => 'no-cache',
	     -pragma => 'no-cache',
	     -expires => 1000000000,
	     -cookie => [cookie(-name => 'webdb',
			       -value => $ENV{WEBDB_SECRET},
			       -path => '/',
			       -expires => '+1d')]);

my $tmpl =
    new MiniTmpl(file => "../template/controller.html");

my $sock;
my %state;

die "WEBDB_PORT not defined" unless defined($ENV{WEBDB_PORT});
my $wait_until = time() + (defined(param('sleep')) ? param('sleep') : 10);
while (time() < $wait_until) {
    $sock = IO::Socket::INET->new(PeerAddr => "localhost:$ENV{WEBDB_PORT}");
    if ($sock) {
	last;
    }
    sleep(1);
}

if (defined($sock)) {
    my $qstr = "cmd=" . $cmd . "&stacklevel=$stacklevel";
    if (param('file') ne '') {
	$qstr .= "&file=" . param('file');
    }
    if (param('line') ne '') {
	$qstr .= "&line=" . param('line');
    }
    if ($cmd eq 'set_breakpoint') {
	$qstr .= "&cond=" . param('cond');
    }
    if (param('set_value') ne '') {
	$qstr .= "&set_value=" . uri_escape(param('set_value'));
    }
    $sock->print("GET /?$qstr HTTP/1.0\r\n\r\n");
    my $status = $sock->getline();
    if ($status =~ m|^HTTP/1.\d+ 200 |) {
	while (1) {
	    my $l = $sock->getline();
	    unless ($l =~ s/[\r\n]+$//) {
		die "unexpected response, got: $l\n";
	    }
	    if ($l eq '') {
		last;
	    }
	}
	my @target = (\%state);
	while (1) {
	    my $l = $sock->getline();
	    unless ($l =~ s/[\r\n]+$//) {
		last;
	    }
	    if ($l =~ /\s*:\(\s*/) {
		my $new_target = $target[0]->{uri_unescape($`)} = {};
		unshift(@target, $new_target);
	    } elsif ($l =~ /^\)/) {
		shift(@target);
	    } elsif ($l =~ /\s*:\s*(.*)/) {
		$target[0]->{uri_unescape($`)} = uri_unescape($1);
	    }
	}
	$sock->close();
        if ($state{terminated}) {
            undef $sock;
        }
    } else {
	$sock->close();
	undef $sock;
   }
}

if ($sock) {
    my $cursor_file;
    { # setup source window information
	my $cur_stack = $state{stack}->{$stacklevel};
	$cursor_file = $cur_stack->{file};
	$tmpl->param(cursor_file => $cursor_file,
		     cursor_line => $cur_stack->{line} + 0);
	if ($cur_stack->{file} eq $state{stack}->{0}->{file}) {
	    $tmpl->param(active_line => $state{stack}->{0}->{line});
	} else {
	    $tmpl->param(active_line => 0);
	}
	my @backtrace_lines;
	for (my $i = 1; defined($state{stack}->{$i}); $i++) {
	    if ($state{stack}->{$i}->{file} eq $cur_stack->{file}) {
		push(@backtrace_lines, $state{stack}->{$i}->{line});
	    }
	}
	$tmpl->param(backtrace_lines => \@backtrace_lines);
    }
    { # setup backtrace list
	my @backtrace_list;
	my %labels;
	foreach my $level (sort(keys(%{$state{stack}}))) {
	    my $frame = $state{stack}->{$level};
	    push(@backtrace_list, { level => $level,
				    package => $frame->{package},
				    function => $frame->{function},
				    file => basename($frame->{file}),
				    line => $frame->{line},
				    selected => ($level == $stacklevel) });
	}
	$tmpl->param(backtrace_list => \@backtrace_list);
    }
    { # setup breakpoints
	my @breakpoints;
	if (my $linecond_hashref = $state{breakpoints}->{$cursor_file}) {
	    foreach my $line_no (sort(keys(%$linecond_hashref))) {
		push(@breakpoints, $line_no);
	    }
	}
	$tmpl->param(breakpoints => \@breakpoints);
    }
    # setup info
    if (param('info') ne 'stderr' &&
	($cmd eq '' || $state{got_stderr} == 0)) {
	my @vars;
	foreach my $type (sort(keys(%{$state{var}}))) {
	    if ($invisible_var_type =~ /(^|,)$type($|,)/) {
		push(@vars, { type => $type, visible => 0 });
	    } else {
		my %var_type = %{$state{var}->{$type}};
		remove_special_vars(\%var_type);
		my @expand = param('var_expand');
		my ($data) = dump_vars(99, \%var_type, "$type", \@expand);
		push(@vars, { type => $type, visible => 1, data => $data });
	    }
	}
	$tmpl->param(var_selected => 1, vars => \@vars);
    } else {
	$tmpl->param(stderr_selected => 1, stderr => $state{stderr});
    }
}

my $window_height = param('height');
if ($window_height < 40) {
    $window_height = 158;
}
$tmpl->param(window_height => $window_height,
	     info_height => $window_height - 36);
$tmpl->param(connected => $sock,
	     running => ($sock && ! $state{terminated}) ? 1 : 0);

print $tmpl->output();


sub dump_vars {
    my ($cols, $h, $path, $expand_vars) = @_;
    my $content = '';
    my $rows = 0;
    foreach my $n (sort(keys(%{$h}))) {
	my $v = $h->{$n};
	my ($subcontent, $subrows);
	my ($addcontrol, $addtag) = ('', 1);
	if (ref($v) eq 'HASH') {
	    $addcontrol = '+';
	    foreach my $patt (@{$expand_vars}) {
		if ($patt eq "$path/$n") {
		    $addcontrol = '-';
		    last;
		}
	    }
	    if ($addcontrol eq '-') {
		if (%{$v} <=> 0) {
		    ($subcontent, $subrows) =
			dump_vars($cols - 1, $v, "$path/$n", $expand_vars);
		    $addtag = 0;
		} else {
		    $subcontent = '';
		    $subrows = 1;
		}
	    } else {
		$subcontent = '...';
		$subrows = 1;
	    }
	} else {
	    my $true_v = $v;
	    $true_v =~ s/\\c(.)/chr(ord($1)^64)/eg;
	    $true_v =~ s/\\e/\033/g;
	    $true_v =~ s/\\(.)/$1/g;
	    $subcontent = escapeHTML($true_v);
	    $subcontent =~ s/\n/<br>/g;
	    $subcontent =~ s/ /&nbsp;/g;
	    $subcontent =~
		s/\t/&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;/g;
	    $subcontent .=
		sprintf(<<'EOF'
&nbsp;&nbsp;<a href="javascript:edit_variable('%s', '%s')"><img src="images/edit_var.gif" width="12" height="12" border="0" alt="Edit Variable"></a>
EOF
			, "$path/$n", js_escape($true_v));
	    $subrows = 1;
	}
	if ($addtag) {
	    $subcontent =
		"<td colspan=\"$cols\" width=\"99%\">$subcontent</td></tr>\n";
	}
	if ($content ne '') {
	    $content .= '<tr>';
	}
	$content .= <<"EOF";
<td nowrap rowspan="$subrows" valign="top">
<a name="var/$path/$n">$n</a>
EOF
;
	if ($addcontrol eq '+') {
	    $content .= <<"EOF";
<a href="javascript:expand_value('$path/$n', true)"><img src="images/open_var.gif" alt="Show Variables" width="10" height="10" border="0"></a>
EOF
;
	} elsif ($addcontrol eq '-') {
	    $content .= <<"EOF";
<input type="hidden" name="var_expand" value="$path/$n">
<a href="javascript:document.theForm.submit()" onClick="set_input_value('var_expand', '$path/$n', ''); adjust_height()"><img src="images/close_var.gif" alt="Hide Variables" width="10" height="10" border="0"></a>
EOF
;
	}
	$content .= "</td>$subcontent";
	$rows += $subrows;
    }
    return ($content, $rows);
}

sub remove_special_vars {
    my ($vars) = @_;
    foreach my $k (keys(%{$vars})) {
	if (($k =~ /^\$(.)/ && $1 !~ /[A-Za-z_]/) ||
	    $k =~ /(\%\!|\%ENV|\%INC|\%SIG|\@\+|\@\-|\@INC)/) {
	    delete $vars->{$k};
	}
    }
}
